#pragma once
///@file

#include "nix/expr/eval.hh"

#include <tuple>
#include <vector>

namespace nix {

struct RegisterPrimOp
{
    typedef std::vector<PrimOp> PrimOps;

    static PrimOps & primOps();

    /**
     * You can register a constant by passing an arity of 0. fun
     * will get called during EvalState initialization, so there
     * may be primops not yet added and builtins is not yet sorted.
     */
    RegisterPrimOp(PrimOp && primOp);
};

/* These primops are disabled without enableNativeCode, but plugins
   may wish to use them in limited contexts without globally enabling
   them. */

/**
 * Load a ValueInitializer from a DSO and return whatever it initializes
 */
void prim_importNative(EvalState & state, const PosIdx pos, Value ** args, Value & v);

/**
 * Execute a program and parse its output
 */
void prim_exec(EvalState & state, const PosIdx pos, Value ** args, Value & v);

void makePositionThunks(EvalState & state, const PosIdx pos, Value & line, Value & column);

} // namespace nix
